failure = object()


class Result:
    """
    Wraps a callable and calls it with the given arguments
    if an exception is raised in the callable, it is wrapped.
    """

    def __init__(self, _callable, *args, suppress=Exception, **kwargs):
        """
        >>> w = Result(lambda x: x, 1)
        >>> w.unwrap()
        1
        >>> w = Result(lambda x: x, 1, suppress=TypeError)
        >>> w.unwrap()
        1
        >>> w = Result(lambda x, y: x, 1)
        >>> w.unwrap()
        Traceback (most recent call last):
        ...
        TypeError: missing required positional argument: 'y'
        >>> w = Result(lambda x, y: x, 1)
        >>> w.unwrap_or("default")
        'default'
        >>> w.unwrap_or_else(lambda: "default")
        'default'
        >>> w.is_ok()
        False
        >>> w.is_error()
        True
        >>> w = Result(lambda: 2)
        >>> w.and_then(lambda value, x: x * value, 5)
        10
        """
        self.exc = None
        self.value = failure
        try:
            self.value = _callable(*args, **kwargs)
        except suppress as e:
            self.exc = e

    def unwrap(self):
        """
        If an exception was raised in the wrapped callable,
        raise it otherwise return the result value.
        """
        if self.exc is not None:
            raise self.exc
        return self.value

    def unwrap_or(self, default):
        """
        If an exception was raised in the wrapped callable,
        return the default value otherwise return the result value.
        """
        return self.value if self.value is not failure else default

    def unwrap_or_else(self, _callable, *args, **kwargs):
        """
        If an exception was raised in the wrapped callable,
        return the result of passed calleble otherwise return the result value.
        """
        return self.unwrap_or(_callable(*args, **kwargs))

    def is_ok(self):
        """
        Tells if the wrapped callable returned without error.
        """
        return self.value is not failure

    def is_error(self):
        """
        Tells if the wrapped callable raised an exception.
        """
        return self.value is failure

    def ok(self):
        """
        Same as unwrap().
        """
        return self.unwrap()

    def and_then(self, _callable, *args, **kwargs):
        """
        If no exception was raised in the wrapped callable,
        pass the result value to the given callable.
        as a wrapped callable.
        """
        return Result(_callable, self.unwrap(), *args, **kwargs)



# borrowed from: dhilst/geckones_twitter_bots/blob/master/utils.py
import contextlib
from typing import Mapping, Sequence


class get:  # noqa
    """
    This will emulate .get method for

    dicts, lists and tuples

    You can use like this
    >>> [1] >> get(2, 'Not found')
    'Not found'
    >>> [1] >> get(2)
    >>> [1] >> get(0)
    1
    >>> {} >> get('a', 'Not found')
    >>> () >> get(0, 'Not found')
    """

    def __init__(self, item, default=None):
        self.item = item
        self.default = default

    def __rrshift__(self, other):
        other_type = get_base_type(other)
        return getattr(self, f"_get_{other_type}", self._get)(other)

    def _get(self, other):
        if hasattr(other, "get"):
            return other.get(self.item, self.default)
        elif hasattr(other, "__getitem__"):
            with contextlib.suppress(KeyError):
                return other[self.item]
            return self.default
        return getattr(other, self.item, self.default)

    def _get_sequence(self, other):
        with contextlib.suppress(IndexError):
            return other[self.item]
        return self.default

    def _get_mapping(self, other):
        return other.get(self.item, self.default)


def get_base_type(other):
    """
    >>> get_base_type(1)
    int
    >>> get_base_type([])
    sequence
    >>> get_base_type({})
    mapping
    """
    if isinstance(other, Mapping):
        return "mapping"
    if isinstance(other, Sequence):
        return "sequence"
    return type(other).__name__.lower()



"""Python setup.py for smarter package"""

from setuptools import find_packages, setup

setup(
    name="smarter",
    version="0.1.1",
    description="Some objects could be smarter",
    author="Bruno Rocha",
    packages=find_packages(exclude=["tests"]),
    install_requires=[],
    extras_require={
        "test": [
            "pytest",
            "coverage",
            "flake8",
            "black",
            "isort",
            "pytest-cov",
            "codecov",
            "mypy",
            "gitchangelog",
        ],
    },
)



from collections import UserList
from typing import Any, Optional


class SmartList(UserList):
    """
    A list with more utility methods added.
    """

    def get(self, index: int, default: Optional[Any] = None) -> Any:
        """
        Return the element at index or default if index is out of range.
        """
        try:
            return self[index]
        except IndexError:
            return default

    def first(self):
        """
        Return first item of the list
        """
        return self[0]

    def last(self):
        """
        Return last item of the list
        """
        return self[-1]

    def first_or(self, default: Any = None):
        """
        Return first item of the list or default if the list is empty.
        """
        return self.get(0, default)

    def last_or(self, default: Any = None):
        """
        Return last item of the list or default if the list is empty.
        """
        return self.get(-1, default)

    def first_not_null(self, default: Optional[Any] = None) -> Any:
        """
        Return the first non-null item of the list or default.
        """
        for item in self:
            if item is not None:
                return item
        return default

    def first_not_nullable(self, default: Optional[Any] = None):
        """
        Return the first non-nullable item of the list or default.
        """
        for item in self:
            if item:
                return item
        return default
